/**
 * Copyright (C) 2008-2012 AgroSense Foundation.
 *
 * AgroSense is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * There are special exceptions to the terms and conditions of the GPLv3 as it is applied to
 * this software, see the FLOSS License Exception
 * <http://www.agrosense.eu/foss-exception.html>.
 *
 * AgroSense is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with AgroSense.  If not, see <http://www.gnu.org/licenses/>.
 */
package eu.agrosense.client.calendar.api.gui;

import static eu.agrosense.client.calendar.api.internal.util.DateUtil.DAY_LONG_FORMAT;
import static eu.agrosense.client.calendar.api.internal.util.DateUtil.DAY_MEDIUM_FORMAT;
import static eu.agrosense.client.calendar.api.internal.util.DateUtil.DAY_SHORT_FORMAT;
import static eu.agrosense.client.calendar.api.internal.util.DateUtil.HOUR_FORMAT;
import static eu.agrosense.client.calendar.api.internal.util.DateUtil.MONTH_LONG_FORMAT;
import static eu.agrosense.client.calendar.api.internal.util.DateUtil.MONTH_MEDIUM_FORMAT;
import static eu.agrosense.client.calendar.api.internal.util.DateUtil.MONTH_SHORT_FORMAT;

import java.lang.reflect.Field;
import java.util.ArrayList;
import java.util.List;

import javafx.beans.value.ChangeListener;
import javafx.collections.ObservableList;
import javafx.event.EventHandler;
import javafx.geometry.Side;
import javafx.scene.Cursor;
import javafx.scene.Node;
import javafx.scene.chart.Axis;
import javafx.scene.input.MouseEvent;
import javafx.scene.text.Text;

import org.joda.time.DateTime;
import org.joda.time.Period;
import org.joda.time.ReadablePeriod;

import eu.agrosense.client.calendar.api.event.AxisEvent;
import eu.agrosense.client.calendar.api.event.AxisEventHandler;
import eu.agrosense.client.calendar.api.impl.gui.LinearAxisCalculator;

/**
 * Date axis, working with DateTime. <br> Time value is needed for precise
 * planning
 *
 * @author post
 *
 */
public class CalendarDateAxis extends Axis<DateTime> {

    private static final int YEAR_INTERVAL = 10;
    private static final int MONTH_INTERVAL = 9;
    private static final int DAY_INTERVAL = 8;
    private static final int HOUR_INTERVAL = 7;
    private static final DateTime DATETIME_ZERO = new DateTime(0);
    
    //TODO magic numbers
    private static final int MONTH_LONG_FORMAT_TRESHOLD = 90;
    private static final int MONTH_SHORT_FORMAT_TRESHOLD = 50;
    private static final int DAY_LONG_FORMAT_TRESHOLD = 120;
    private static final int DAY_SHORT_FORMAT_TRESHOLD = 70;
    
    private DateTime minDate;
    private DateTime maxDate;
    private ChangeListener<Number> changeListener;
    private AxisEventHandler<DateTime> axisEventHandler;
    private EventHandler<MouseEvent> tickMarkEventHandler;
	private boolean rangeChanged = false;
	private double lastAxisWidth = 0;
	
	private AxisCalculator axisCalculator;

    public CalendarDateAxis(DateTime minDate, DateTime maxDate) {
        this.minDate = minDate;
        this.maxDate = maxDate;
        initListeners();
        axisCalculator = new LinearAxisCalculator();
    }

	@Override
    protected Object autoRange(double length) {
        return new DateTime[]{minDate, maxDate};
    }

    @Override
    protected List<DateTime> calculateTickValues(double length, Object range) {

        final DateTime[] rangeProps = (DateTime[]) range;
        final DateTime lowerBound = rangeProps[0];
        final DateTime upperBound = rangeProps[1];
        List<DateTime> tickValues = new ArrayList<>();
        if (lowerBound == upperBound) {
            tickValues.add(lowerBound);
        } else {
            ReadablePeriod p = getTickUnit(lowerBound, upperBound);
            DateTime startDate = new DateTime(lowerBound);
            startDate = moveDateToNearestFullValue(startDate);
            
            long max = new DateTime(upperBound).toDate().getTime();
            for (DateTime date = startDate; date.toDate().getTime() <= max; date = date.plus(p)) {
                tickValues.add(new DateTime(date.toDate().getTime()));
            }
        }
        return tickValues;
    }
    
    //wee need to paint ticks at the beginning of period, no at the beginning of the graph (start of the day, week, month)
    private DateTime moveDateToNearestFullValue(DateTime dateTime) {
    	int interval = getInterval(minDate, maxDate);
    	DateTime output;
    	
        switch (interval) {
            case YEAR_INTERVAL:
                output = dateTime.withDayOfMonth(1);
                break;
            case MONTH_INTERVAL:
                output = dateTime.withDayOfMonth(1).toDateMidnight().toDateTime();
                break;
            case DAY_INTERVAL: 
                output = dateTime.toDateMidnight().toDateTime();
                break;
            case HOUR_INTERVAL:
                output = dateTime.withMinuteOfHour(0).withSecondOfMinute(0).withMillisOfSecond(0);
                break;
            default:
            	 output = dateTime.withDayOfMonth(1);
                break;
        }
        
        return output;
    }

    //if interval is year, than ticks are months etc.
    private ReadablePeriod getTickUnit(DateTime lowerBound, DateTime upperBound) {
        int interval = getInterval(lowerBound, upperBound);
        Period p = new Period();

        switch (interval) {
            case YEAR_INTERVAL:
                p = p.plusMonths(1);
                break;
            case MONTH_INTERVAL:
                p = p.plusDays(1);
                break;
            case DAY_INTERVAL:
                p = p.plusDays(1);
                break;
            case HOUR_INTERVAL:
                p = p.plusHours(1);
                break;
            default:
                p = p.plusMonths(1);
                break;
        }

        return p;
    }

    private int getInterval(DateTime lowerBound, DateTime upperBound) {
        long intervalLength = upperBound.getMillis() - lowerBound.getMillis();
        return (int) Math.floor(Math.log10(intervalLength));
    }

    private String formatDate(DateTime dateTime) {
        int interval = getInterval(minDate, maxDate);
        ReadablePeriod period = getTickUnit(minDate, maxDate);
        String format;
        switch (interval) {
            case YEAR_INTERVAL:
            	format = getMonthFormat(dateTime, period);
            	break;
            case MONTH_INTERVAL:
            	format = getDayFormat(dateTime, period);
            	break;
            case DAY_INTERVAL:
            	format =  getDayFormat(dateTime, period);
            	break;
            case HOUR_INTERVAL:
            	format =  getHourFormat();
            	break;
            default:
            	format = getMonthFormat(dateTime, period);
            	break;
        }
        return dateTime.toString(format);
    }

    private double getTickWidth(DateTime dateTime, ReadablePeriod period) {
    	DateTime dateTimeNext = dateTime.plus(period);
    	return (getDisplayPosition(dateTimeNext) - getDisplayPosition(dateTime));
    }
    
    private String getMonthFormat(DateTime dateTime, ReadablePeriod period) {
    	double tickWidth = getTickWidth(dateTime, period);
    	if (tickWidth > MONTH_LONG_FORMAT_TRESHOLD) {
    		return MONTH_LONG_FORMAT;
    	} else if (tickWidth <= MONTH_LONG_FORMAT_TRESHOLD && tickWidth > MONTH_SHORT_FORMAT_TRESHOLD){
    		return MONTH_MEDIUM_FORMAT;
    	} else {
    		return MONTH_SHORT_FORMAT;
    	}
    }
    
    private String getDayFormat(DateTime dateTime, ReadablePeriod period) {
    	double tickWidth = getTickWidth(dateTime, period);
    	
    	if (tickWidth > DAY_LONG_FORMAT_TRESHOLD) {
    		return DAY_LONG_FORMAT;
    	} else if (tickWidth <= DAY_LONG_FORMAT_TRESHOLD && tickWidth > DAY_SHORT_FORMAT_TRESHOLD){
    		return DAY_MEDIUM_FORMAT;
    	} else {
    		return DAY_SHORT_FORMAT;
    	}
    }
    
    private String getHourFormat() {
    	return HOUR_FORMAT;
    }
    
    @Override
    public double getDisplayPosition(DateTime date) {
        double length = (Side.TOP.equals(getSide()) || Side.BOTTOM.equals(getSide())) ? getWidth() : getHeight();
        return axisCalculator.getDisplayPosition(date, length, minDate, maxDate);
    }

    @Override
    protected Object getRange() {
        return new Object[]{minDate, maxDate};
    }

    @Override
    protected String getTickMarkLabel(DateTime date) {
        return formatDate(date);
    }

    @Override
    public DateTime getValueForDisplay(double value) {
        double length = (Side.TOP.equals(getSide()) || Side.BOTTOM.equals(getSide())) ? getWidth() : getHeight();
        return axisCalculator.getValueForDisplay(value, length, minDate, maxDate);
    }

    @Override
    public double getZeroPosition() {
        return 0;
    }

    @Override
    public boolean isValueOnAxis(DateTime value) {
        return (value.isAfter(minDate) || value.equals(minDate)) && (value.isBefore(maxDate) || value.equals(maxDate));
    }

    @Override
    public void setRange(Object range, boolean animate) {
        Object[] r = (Object[]) range;
        DateTime minDateNew = (DateTime) r[0];
        DateTime maxDateNew = (DateTime) r[1];
        if (!minDateNew.equals(minDate) || !maxDateNew.equals(maxDate)) {
        	rangeChanged = true;
        	minDate = minDateNew;
        	maxDate = maxDateNew;
        	long rangeValue = maxDate.getMillis() - minDate.getMillis();
        	fireRangeChange(0, rangeValue);
        }
        
    }

    @Override
    public double toNumericValue(DateTime dateTime) {
        return dateTime.getMillis();
    }

    @Override
    public DateTime toRealValue(double v) {
        return new DateTime((long) v);
    }

    //overriden to change position of axis tick labels
    @Override
    protected void layoutChildren() {
    	//when changing range from year to month, tickMark corresponding to beginning of the month is the same as tickmark from year range.
    	//because of this, the tickMark is not removed and is reused, but with wrongly formatted value.
    	//we can not remove tickMarks, so we change value of all tickMarks, which will lead to proper remove in super.layoutChildren();
		boolean widthChanged = false;
    	if (lastAxisWidth != getWidth()) {
			widthChanged = true;
		}
		lastAxisWidth = getWidth();
    	if (rangeChanged || widthChanged) {
			for (TickMark<DateTime> tickMark : getTickMarks()) {
				tickMark.setValue(DATETIME_ZERO);
			}
			rangeChanged = false;
		}
    	
        super.layoutChildren();

        List<Text> textList = new ArrayList<>();

        Class<?> clazz = TickMark.class;
        try {
            Field textNodeField = clazz.getDeclaredField("textNode");
            textNodeField.setAccessible(true);

            for (TickMark<DateTime> tick : getTickMarks()) {
                Text textNode = (Text) textNodeField.get(tick);
                textList.add(textNode);
            }

        } catch (NoSuchFieldException | SecurityException | IllegalArgumentException | IllegalAccessException e) {
            e.printStackTrace(); //TODO how to log from FX thread?
        }

        double maxValuePos = getDisplayPosition(maxDate);
        
        ReadablePeriod period = getTickUnit(minDate, maxDate);
        ObservableList<TickMark<DateTime>> tickMarks = getTickMarks();
        
        for (int i = 0; i < textList.size(); i++) {
        	Text text = textList.get(i);
        	TickMark<DateTime> tickMark = tickMarks.get(i);
        	double gap = getTickGap(tickMark.getValue(), period);
        	double tickWidth = 2 * gap -2; //leave some space
        	
                if (!text.isVisible()) {
                    boolean hasEnoughSpace = text.getLayoutBounds().getWidth() < tickWidth;
                    text.setVisible(hasEnoughSpace);
                }
                
        	text.setLayoutX(text.getLayoutX() + gap);
            
            if (lessThanHalfVisible(text, maxValuePos)) {
                text.setVisible(false);
            } 
            
            if (text.isVisible()) {
            	text.setOnMouseClicked(tickMarkEventHandler); 
            	text.setCursor(Cursor.HAND);
            }
        }
    }
    
    private boolean lessThanHalfVisible(Text text, double width) {
    	return text.getLayoutX()  < - (text.getLayoutBounds().getWidth()/2) || text.getLayoutX() + (text.getLayoutBounds().getWidth()/2) > width;
    }

    private double getTickGap(DateTime dateTime, ReadablePeriod period) {
    	return getTickWidth(dateTime, period) / 2d;
    }
    
    public void setOnRangePropertyChangeListener(ChangeListener<Number> changeListener) {
        this.changeListener = changeListener;
    }

    private void fireRangeChange(long oldValue, long newValue) {
        if (changeListener != null) {
            changeListener.changed(null, oldValue, newValue);
        }
    }
    
    public void setOnAxisClickedHandler(AxisEventHandler<DateTime> axisEventHandler) {
    	this.axisEventHandler = axisEventHandler;
    }

    private void initListeners() {
    	tickMarkEventHandler = new EventHandler<MouseEvent>() {
			
			@Override
			public void handle(MouseEvent mouseEvent) {
				mouseClicked(mouseEvent);
			}
			
		};
    }
    
    protected void mouseClicked(MouseEvent mouseEvent) {
		if (axisEventHandler != null) {
			Node node = (Node) mouseEvent.getSource();
			double position = node.getLayoutX() + node.getLayoutBounds().getWidth() / 2;
			DateTime clickedDate = getValueForDisplay(position);
			AxisEvent<DateTime> event = new AxisEvent<>(clickedDate);
			axisEventHandler.handle(event);
		}
	}
    
    public void setAxisCalculator(AxisCalculator axisCalculator) {
    	if (axisCalculator != this.axisCalculator) {
    		rangeChanged = true;
    	}
  		this.axisCalculator = axisCalculator;
	}
    
    public AxisCalculator getAxisCalculator() {
		return axisCalculator;
	}
}
