/**
 * Copyright (C) 2008-2012 AgroSense Foundation.
 *
 * AgroSense is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * There are special exceptions to the terms and conditions of the GPLv3 as it is applied to
 * this software, see the FLOSS License Exception
 * <http://www.agrosense.eu/foss-exception.html>.
 *
 * AgroSense is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with AgroSense.  If not, see <http://www.gnu.org/licenses/>.
 */
package eu.agrosense.client.calendar.api.model;

import java.util.ArrayList;
import java.util.List;

import javafx.beans.value.ObservableValueBase;

import org.joda.time.DateTime;
import org.joda.time.LocalDate;

import eu.agrosense.client.calendar.api.internal.util.DateUtil;

public class CalendarNode extends ObservableValueBase<CalendarNode>{

    private String name;
    private DateTime start;
    private DateTime end;
    private List<CalendarNode> children = new ArrayList<>();
    private CalendarNode parent;
    private boolean visible = true;
    private boolean expanded = true;

    public CalendarNode() {
    	super();
    }
    
    public CalendarNode(String name, DateTime start, DateTime end) {
        super();
        this.name = name;
        this.start = start;
        this.end = end;
    }

    public CalendarNode(String name, LocalDate start, LocalDate end) {
        super();
        this.name = name;
        this.start = DateUtil.startOfDay(start);
        this.end = DateUtil.endOfDay(end);
    }

    public DateTime getStart() {
        return start;
    }

    public DateTime getEnd() {
        return end;
    }

    public CalendarNode getParent() {
        return parent;
    }

    void setParent(CalendarNode parent) {
        this.parent = parent;
    }

    public void addChild(CalendarNode node) {
        if (node.getParent() != null) {
            throw new IllegalStateException("model already has a parent");
        }
        children.add(node);
        node.setParent(this);
        fireValueChangedEvent();
    }

    public void removeChild(CalendarNode node) {
        if (node.getParent() != this) {
            throw new IllegalStateException("doesn't contain this model");
        }
        children.remove(node);
        node.setParent(null);
        fireValueChangedEvent();
    }

    public String getName() {
        return name;
    }

    public void setName(String name) {
        this.name = name;
    }

    //TODO immutable?
    public List<CalendarNode> getChildren() {
        return children;
    }
    
    public void setExpanded(boolean expanded) {
		this.expanded = expanded;
	}
    
    public void setVisible(boolean visible) {
		this.visible = visible;
	}
    
    public boolean isExpanded() {
		return expanded;
	}
    
    public boolean isVisible() {
		return visible;
	}
    
    public boolean isVisibleInHierarchy() {
		if (isVisible() && parent != null) {
			return parent.isVisibleInHierarchy() && parent.isExpanded();
		} else {
			return isVisible();
		}
	}

	@Override
	public int hashCode() {
		final int prime = 31;
		int result = 1;
		result = prime * result
				+ ((children == null) ? 0 : children.hashCode());
		result = prime * result + ((end == null) ? 0 : end.hashCode());
		result = prime * result + ((name == null) ? 0 : name.hashCode());
		result = prime * result + ((start == null) ? 0 : start.hashCode());
		return result;
	}

	@Override
	public boolean equals(Object obj) {
		if (this == obj)
			return true;
		if (obj == null)
			return false;
		if (getClass() != obj.getClass())
			return false;
		CalendarNode other = (CalendarNode) obj;
		if (children == null) {
			if (other.children != null)
				return false;
		} else if (!children.equals(other.children))
			return false;
		if (end == null) {
			if (other.end != null)
				return false;
		} else if (!end.equals(other.end))
			return false;
		if (name == null) {
			if (other.name != null)
				return false;
		} else if (!name.equals(other.name))
			return false;
		if (start == null) {
			if (other.start != null)
				return false;
		} else if (!start.equals(other.start))
			return false;
		return true;
	}

	@Override
	public CalendarNode getValue() {
		return this;
	}

}
